// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit mtmanager;

interface

{$I dorb.inc}

uses
  {$IFDEF MSWINDOWS}Windows,{$ENDIF}{$IFDEF LINUX}Libc,{$ENDIF}SysUtils, Classes,
  Contnrs, mtmanager_int, osthread, orbtypes;

{$IFNDEF USELIB}

type
  TMsgChannelTypes = (mcActive, mcPassive, mcDirect);

  // Forward declarations
  TThreadPool = class;
  TThreadPoolManager = class;
  TWorkerThread = class;

  TMsgType = class(TInterfacedObject, IMsgType)
  protected
    FType: TMTMesType;
    FPtr: pointer;
    FValue: integer;
    function get_type: TMTMesType;
    procedure set_type(msgType: TMTMesType);
    function data: pointer;
  public
    constructor Create(ptr: pointer; AType: TMTMesType = mtProcess);
    destructor Destroy; override;
    property Value: integer read FValue write FValue;
  end;

  TMsgChannel = class(TInterfacedObject, IMsgChannel)
  protected
    FThreadPool: IThreadPool;
    procedure set_thread_pool(AThreadPool: IThreadPool);
    function get_thread_pool: IThreadPool; overload;
    procedure check_msg();
    //function check_msg(var kt: IWorkerThread): boolean; virtual;
    function get_msg(idType: TOperationKind): IMsgType; virtual;
    procedure put_msg(nextOpId: TOperationKind; msg: IMsgType); virtual; abstract;
  public
  end;

  TActiveMsgQueue = class(TMsgChannel, IMsgChannel)
  private
    FQueue: TQueue;
    FAccess: TMutex;
  protected
    procedure check_msg();
    //function check_msg(var kt: IWorkerThread): boolean; override;
    procedure put_msg(nextOpId: TOperationKind; msg: IMsgType); override;
  public
    constructor Create();
    destructor Destroy; override;
  end;

  TPassiveMsgQueue = class(TMsgChannel, IPassiveMsgQueue)
  private
    Fkt: IWorkerThread;
    FQueue: TQueue;
    FAccess: TMutex;
    FMsgCnt: TSemaphore;
  protected
    procedure set_worker_thread(kt: IWorkerThread);
    function get_msg(idType: TOperationKind): IMsgType; override;
    procedure put_msg(idType: TOperationKind; msgType: IMsgType); override;
  public
    constructor Create();
    destructor Destroy; override;
  end;

  TDirectMsgConnector = class(TMsgChannel, IDirectMsgConnector)
  private
    FOp: IOperation;
  protected
    procedure set_operation(op: IOperation);
    procedure put_msg(idType: TOperationKind; msgType: IMsgType); override;
  public
  end;

  TOperation = class(TMsgChannel, IOperation)
  private
  protected
    FInfo: IOperationInfoNode;
    FWThread: IWorkerThread;
    function get_info(): IOperationInfoNode;
    procedure set_info(value: IOperationInfoNode);
    function get_thread(): IWorkerThread;
    procedure set_thread(value: IWorkerThread);
    procedure start();
    procedure init_shutdown(); virtual;
    procedure finalize_shutdown(); virtual;
    procedure send_msg(nextOpId: TOperationKind; msg: IMsgType);
    procedure process(msg: IMsgType); virtual;
    procedure run(); virtual; abstract;
    function copy(): IOperation; virtual; abstract;
    procedure put_msg(nextOpId: TOperationKind; msg: IMsgType); override;
  public
    constructor Create(); overload;
    constructor Create(AThreadPool: IThreadPool); overload;
    constructor Create(op: IOperation); overload;
    destructor Destroy; override;
  end;

  TOperationInfoNode = class(TInterfacedObject, IOperationInfoNode)
  private
    FNextOP: IMsgChannel;
    FTp: IThreadPool;
    FOpID: TOperationKind;
  protected
    function get_thread_pool(): IThreadPool;
    procedure set_thread_pool(tp: IThreadPool);
    function get_OP_id(): TOperationKind;
    procedure set_OP_id(id: TOperationKind);
    function get_nextOP(): IMsgChannel;
    procedure set_nextOP(nextOP: IMsgChannel);
  end;

  TActiveOperation = class(TOperation, IActiveOperation)
  private
  protected
    FInputMC: IPassiveMsgQueue;
    procedure register_inputMC(mc: IPassiveMsgQueue);
    function get_inputMC(): IPassiveMsgQueue;
    procedure init_shutdown(); override;
    procedure run(); override;
  public
    constructor Create(AThreadPool: IThreadPool); overload;
    constructor Create(op: IActiveOperation); overload;
  end;

  TPassiveOperation = class(TOperation, IPassiveOperation)
  private
    FMsgType: IMsgType;
  protected
    procedure put_msg(idType: TOperationKind; msgType: IMsgType); override;
    procedure run(); override;
  end;

  TWorkerThread = class(TORBThread, IWorkerThread)
  private
    FTP: IThreadPool;
    FOp: IOperation;
    FState: TWorkerThreadState;
    FOpCritical: TRecursiveMutex;
    FStateSem: TSemaphore;
    procedure wait_for_state_change();
    function get_state: TWorkerThreadState;
    procedure set_state(const Value: TWorkerThreadState);
  protected
    procedure set_thread_pool(AThreadPool: IThreadPool);
    function get_thread_pool: IThreadPool;
    procedure register_operation(op: IOperation);
    procedure deregister_operation(op: IOperation);
    procedure post_state_change();
    procedure put_msg(msgType: IMsgType);
    procedure mark_idle();
    procedure mark_busy();
    function get_operation(): IOperation;
    procedure Run(); override;
  public
    constructor Create(); overload;
    constructor Create(AThreadPool: IThreadPool); overload;
    destructor Destroy; override;
  end;

  TWorkerThreadList = class(TList);

  TThreadPool = class(TMsgChannel, IThreadPool)
  private
    FOpType: integer;
    //FTPLock: TMutex;
    FTpm: IThreadPoolManager;
    FBusyThreads,
    FIdleThreads: IInterfaceList;//TList;
    FOp: IOperation;
    FInputMC: IMsgChannel;//TMsgChannel;
    FMax,
    FMaxIdle,
    FMinIdle,
    FCountAll: cardinal;
    FMonitor: TThreadMonitor;
    function new_idle_thread(): IWorkerThread;
  protected
    function get_idle_thread(): IWorkerThread;
    procedure mark_busy(wt: IWorkerThread);
    procedure mark_idle(wt: IWorkerThread);
    function remove_thread(kt: IWorkerThread): boolean;
    function get_operation(): IOperation;
    procedure set_operation(op: IOperation);
    procedure set_thread_pool_manager(tpm: IThreadPoolManager);
    function get_thread_pool_manager(): IThreadPoolManager;
    function optype: integer;
    procedure clear();
    // fire up all threads
    procedure start_threads(arg: pointer = nil); virtual;
    procedure registerInputMC(mc: TMsgChannel);
    function getInputMC(): IMsgChannel;//TMsgChannel;
    procedure put_msg(nextOpId: TOperationKind; msgType: IMsgType); override;
  public
    constructor Create(AMax, AMaxIdle, AMinIdle: cardinal);
    destructor Destroy; override;
  end;

  TThreadPoolManager = class(TMsgChannel, IThreadPoolManager)
  private
    FTp: TStringList;
    FUserCnt: _ulong;
    procedure register_tp(const opId: TOperationKind; tp: IThreadPool);
  protected
    function register_new_tp(tp: IThreadPool): _ulong;
    function get_operation(opId: TOperationKind): IOperation;
    procedure put_msg(nextOpId: TOperationKind; msgType: IMsgType); override;
    { IThreadPoolManager }
    function get_thread_pool(const opId: TOperationKind): IThreadPool; overload;
    function get_idle_thread(const opId: TOperationKind): IWorkerThread;
    procedure shutdown();
  public
    constructor Create();
    destructor Destroy; override;
  end;

  TMTDispatcher = class(TPassiveOperation, IPassiveOperation)
  protected
    procedure process(msg: IMsgType); override;
    function copy(): IOperation; override;
  public
    constructor Create(); overload;
    constructor Create(op: IOperation); overload;
  end;

function TPM(): IThreadPoolManager;
procedure thread_setup();
procedure thread_shutdown();

implementation

uses
  iop,giop_int,orb;

type
  TTm_Init_T = record
    opKind: TOperationKind;
    max: cardinal;
    max_idle: cardinal;
    min_idle: cardinal;
    mct: TMsgChannelTypes;
    mc_size: cardinal;
    op: IOperation;
    start: boolean;
  end;

const
  tmInitSize = 5;
  tmInit: array [0..tmInitSize-1] of TTm_Init_T =
   ((opKind: odAccept; max: 20; max_idle: 0; min_idle: 0; mct: mcPassive; mc_size: 0; op: nil; start: FALSE),
    (opKind: odWriter; max: 20; max_idle: 0; min_idle: 0; mct: mcPassive; mc_size: 0; op: nil; start: FALSE),
    (opKind: odReader; max: 20; max_idle: 0; min_idle: 0; mct: mcPassive; mc_size: 0; op: nil; start: FALSE),
    (opKind: odDeCode; max: 20; max_idle: 0; min_idle: 0; mct: mcActive;  mc_size: 1; op: nil; start: TRUE),
    (opKind: odORB;    max: 10; max_idle: 0; min_idle: 0; mct: mcActive;  mc_size: 1; op: nil; start: TRUE));

var
  tpmVar: IThreadPoolManager;
  
function TPM(): IThreadPoolManager;
begin
  result := tpmVar;
end;

procedure thread_setup();
var
  i: integer;
  tpmObj: TThreadPoolManager;
  tpObj: TThreadPool;
  tp: IThreadPool;
  mq: TDirectMsgConnector;
begin
  tpmObj := TThreadPoolManager.Create;
  for i := 0 to tmInitSize-1 do begin
    tpObj := TThreadPool.Create(tmInit[i].max, tmInit[i].max_idle, tmInit[i].min_idle);
    tp := tpObj; tpObj.FOpType := ord(tmInit[i].opKind);
    tpmObj.register_tp(tmInit[i].opKind, tp);
    case tmInit[i].opKind of
      odDeCode: tpObj.set_operation(TInputHandler.Create() as IPassiveOperation);
      odORB: tpObj.set_operation(TMTDispatcher.Create() as IPassiveOperation);
    end;
    case tmInit[i].mct of
      mcActive:
        if tmInit[i].mc_size <> 0 then tpObj.registerInputMC(TActiveMsgQueue.Create());
      mcDirect:
        begin
          mq := TDirectMsgConnector.Create;
          mq.set_operation(tpObj.get_operation());
          tpObj.registerInputMC(mq);
        end;
    end;
    if tmInit[i].start then tpObj.start_threads();
  end;
  tpmVar := tpmObj;
end;

procedure thread_shutdown();
begin
  tpmVar.shutdown();
end;

{ TOperationInfoNode }

function TOperationInfoNode.get_nextOP: IMsgChannel;
begin
  result := FNextOP;
end;

function TOperationInfoNode.get_OP_id: TOperationKind;
begin
  result := FOpID;
end;

function TOperationInfoNode.get_thread_pool: IThreadPool;
begin
  result := FTp;
end;

procedure TOperationInfoNode.set_nextOP(nextOP: IMsgChannel);
begin
  FNextOP := nextOP;
end;

procedure TOperationInfoNode.set_OP_id(id: TOperationKind);
begin
  FOpID := id;
end;

procedure TOperationInfoNode.set_thread_pool(tp: IThreadPool);
begin
  FTp := tp;
end;

{ TMsgChannel }

//function TMsgChannel.check_msg(var kt: IWorkerThread): boolean;
procedure TMsgChannel.check_msg();
begin
  //result := false;
end;

function TMsgChannel.get_msg(idType: TOperationKind): IMsgType;
begin
  Assert(false, 'Not implemented');
  result := nil;
end;

function TMsgChannel.get_thread_pool: IThreadPool;
begin
  result := FThreadPool;
end;

procedure TMsgChannel.set_thread_pool(AThreadPool: IThreadPool);
begin
  FThreadPool := AThreadPool;
end;

{ TOperation }

constructor TOperation.Create(AThreadPool: IThreadPool);
begin
  FInfo := TOperationInfoNode.Create();
  FInfo.set_thread_pool(AThreadPool);
end;

constructor TOperation.Create;
begin
  FInfo := TOperationInfoNode.Create();
end;

constructor TOperation.Create(op: IOperation);
begin
  FInfo := op.get_info();
end;

destructor TOperation.Destroy;
begin
  FInfo := nil;
  inherited;
end;

procedure TOperation.finalize_shutdown;
begin
  // make sure the thread has left the context of
  // this OP, deregisterOP will take care of that !
  if FWThread <> nil then FWThread.deregister_operation(Self as IOperation);
end;

function TOperation.get_info: IOperationInfoNode;
begin
  result := FInfo;
end;

function TOperation.get_thread: IWorkerThread;
begin
  result := FWThread;
end;

procedure TOperation.init_shutdown;
begin

end;

// the assert's below mark methods that have to be overloaded
// in derived Classes
// the derived Classes have to overload only one of those !!!
procedure TOperation.process(msg: IMsgType);
begin
  Assert(false, 'Implementation error - this should never happen');
end;

procedure TOperation.put_msg(nextOpId: TOperationKind; msg: IMsgType);
begin
  Assert(false, 'not implemented');
end;

procedure TOperation.send_msg(nextOpId: TOperationKind; msg: IMsgType);
begin
  FInfo.get_nextOP().put_msg(nextOpId, msg);
end;

procedure TOperation.set_info(value: IOperationInfoNode);
begin
  FInfo := value;
end;

procedure TOperation.set_thread(value: IWorkerThread);
begin
  FWThread := value;
end;

procedure TOperation.start;
begin
  FWThread.post_state_change();
end;

{ TActiveOperation }

procedure TActiveOperation.run;
var
  msg: IMsgType;
begin
  while true do begin
    msg := get_inputMC().get_msg(FInfo.get_OP_id());
    // We are guaranteed that msg is not NULL because get_msg() asserts
    // this, so don't bother checking it again after the call to get_msg()
    // above!!!
    if msg.get_type() = mtTerminate then begin
      msg := nil;
      Exit;
    end;
    process(msg);
  end; { while }
end;

constructor TActiveOperation.Create(AThreadPool: IThreadPool);
begin
  inherited Create(AThreadPool);
end;

constructor TActiveOperation.Create(op: IActiveOperation);
begin
  inherited Create(op);
  FInputMC := op.get_inputMC();
end;

function TActiveOperation.get_inputMC: IPassiveMsgQueue;
begin
  result := FInputMC;
end;

procedure TActiveOperation.init_shutdown;
var
  msg: IMsgType;
begin
  if get_inputMC() <> nil then begin
    msg := TMsgType.Create(nil, mtTerminate);
    get_inputMC().put_msg(FInfo.get_OP_id(), msg)
  end;
end;

procedure TActiveOperation.register_inputMC(mc: IPassiveMsgQueue);
begin
  FInputMC := mc;
  FInputMC.set_worker_thread(FWThread);
end;

  { TPassiveOperation }

procedure TPassiveOperation.run;
begin
  if (FMsgType = nil) or (FMsgType.get_type() = mtTerminate) then begin
    FMsgType := nil;
    Exit;
  end;
  process(FMsgType);
  FMsgType := nil;
end;

procedure TPassiveOperation.put_msg(idType: TOperationKind;
  msgType: IMsgType);
begin
  Assert(Finfo.get_OP_id = idType);
  FMsgType := msgType;
end;

  { TMsgType }

constructor TMsgType.Create(ptr: pointer; AType: TMTMesType);
begin
  FPtr := ptr;
  FType := AType;
end;

function TMsgType.data: pointer;
begin
  result := FPtr;
end;

destructor TMsgType.Destroy;
begin
  inherited;
end;

function TMsgType.get_type: TMTMesType;
begin
  result := FType;
end;

procedure TMsgType.set_type(msgType: TMTMesType);
begin
  FType := msgType;
end;

{ TActiveMsgQueue }

//function TActiveMsgQueue.check_msg(var kt: IWorkerThread): boolean;
procedure TActiveMsgQueue.check_msg();
var
  kt: IWorkerThread;
  msg: IMsgType;
begin
  FAccess.lock();
  try
    //result := FQueue.Count <> 0;
    if {result} FQueue.Count <> 0 then begin
      if kt = nil then begin
        kt := FThreadPool.get_idle_thread();
        if kt = nil then begin
          //result := false;
          Exit;
        end;
      end;
      msg := IMsgType(FQueue.Peek);
      Assert(msg <> nil);
      FQueue.Pop;
      kt.put_msg(msg);
      kt.mark_busy();
      kt.post_state_change();
    end;
  finally
    FAccess.unlock();
  end; { try/finally }
end;

constructor TActiveMsgQueue.Create;
begin
  FQueue := TQueue.Create;
  FAccess := TMutex.Create;
end;

destructor TActiveMsgQueue.Destroy;
begin
  FAccess.Free;
  FQueue.Free;
  inherited;
end;

procedure TActiveMsgQueue.put_msg(nextOpId: TOperationKind; msg: IMsgType);
var
  kt: IWorkerThread;
begin
  Assert(msg <> nil);
  FAccess.lock();
  try
    kt := FThreadPool.get_idle_thread();
    if kt <> nil then begin
      kt.put_msg(msg);
      kt.mark_busy;
      kt.post_state_change;
    end
    else begin
        msg._AddRef;
        FQueue.Push(Pointer(msg));
    end;
  finally
    FAccess.unlock();
  end; { try/finally }
end;

  { TPassiveMsgQueue }

constructor TPassiveMsgQueue.Create;
begin
  FQueue := TQueue.Create;
  FAccess := TMutex.Create;
  FMsgCnt := TSemaphore.Create;
end;

destructor TPassiveMsgQueue.Destroy;
begin
  FMsgCnt.Free;
  FAccess.Free;
  FQueue.Free;
  inherited;
end;

function TPassiveMsgQueue.get_msg(idType: TOperationKind): IMsgType;
begin
  FMsgCnt.wait(); // wait for a message (block)
  FAccess.lock();
  try
    result := IMsgType(FQueue.Peek);
    Assert(result <> nil);
    result._Release;
    FQueue.Pop;
  finally
    FAccess.unlock();
  end; { try/finally }
end;

procedure TPassiveMsgQueue.put_msg(idType: TOperationKind; msgType: IMsgType);
begin
  Assert(msgType <> nil);
  FAccess.lock();
  try
    msgType._AddRef;
    FQueue.Push(Pointer(msgType));
  finally
    FAccess.unlock();
  end; { try/finally }
  FMsgCnt.post();
end;

procedure TPassiveMsgQueue.set_worker_thread(kt: IWorkerThread);
begin
  FKt := kt;
end;

{ TDirectMsgConnector }

procedure TDirectMsgConnector.put_msg(idType: TOperationKind;
  msgType: IMsgType);
var
  oper: IOperation;
begin
  Assert(msgType <> nil);
  oper := FOp.copy();
  oper.put_msg(idType, msgType);
  oper.run();
  oper := nil;
end;

procedure TDirectMsgConnector.set_operation(op: IOperation);
begin
  FOp := op;
end;

{ TWorkerThread }

procedure TWorkerThread.Run();
var
  op: IOperation;
begin
  repeat
    wait_for_state_change();
    if FState = wtsBusy then begin
      FOpCritical.lock();
      try
        op := get_operation();
        if op <> nil then op.run();
        op := nil;
      finally
        FOpCritical.unlock();
      end; { try/finally }
      mark_idle();
      post_state_change();
    end;
  until FState = wtsStop;
end;

{procedure TWorkerThread._run_multi_queue_input;
begin

end;

procedure TWorkerThread._run_process;
begin

end;

procedure TWorkerThread._run_single_queue_input;
begin

end;}

constructor TWorkerThread.Create;
begin
  //create mutex and semaphore before inherited Create!!!
  FOpCritical := TRecursiveMutex.Create();
  FStateSem := TSemaphore.Create();
  inherited Create();
end;

constructor TWorkerThread.Create(AThreadPool: IThreadPool);
begin
  Create();
  FTP := AThreadPool;
end;

procedure TWorkerThread.deregister_operation(op: IOperation);
begin
  FOpCritical.lock();
  try
    if op = FOp then begin
      FOp.set_info(nil);
      FOp := nil;
    end;
  finally
    FOpCritical.unlock();
  end; { try/finally }
end;

destructor TWorkerThread.Destroy;
begin
  FRefCount := 1;
  FTP.remove_thread(Self);
  FOpCritical.Free;
  FStateSem.Free;
  inherited;
end;

function TWorkerThread.get_operation: IOperation;
begin
  result := FOp;
end;

function TWorkerThread.get_state: TWorkerThreadState;
begin
  result := FState;
end;

function TWorkerThread.get_thread_pool: IThreadPool;
begin
  result := FTP;
end;

procedure TWorkerThread.mark_busy;
begin
  if get_thread_pool() <> nil then get_thread_pool().mark_busy(Self);
end;

procedure TWorkerThread.mark_idle;
begin
  if get_thread_pool() <> nil then get_thread_pool().mark_idle(Self);
end;

procedure TWorkerThread.post_state_change;
begin
  FStateSem.post();
  //FStateMon.notify();
end;

procedure TWorkerThread.put_msg(msgType: IMsgType);
begin
  FOp.put_msg(FOp.get_info().get_OP_id(), msgType)
end;

procedure TWorkerThread.register_operation(op: IOperation);
begin
  FOpCritical.lock();
  try
    if op <> nil then begin
      FOp := op;
      FOp.set_thread(Self);
    end
    else if FOp <> nil then begin
      FOp.set_thread(nil);
      FOp := nil;
    end;
  finally
    FOpCritical.unlock();
  end; { try/finally }
end;

procedure TWorkerThread.set_state(const Value: TWorkerThreadState);
begin
  FState := Value;
end;

procedure TWorkerThread.set_thread_pool(AThreadPool: IThreadPool);
begin
  FTP := AThreadPool;
end;

procedure TWorkerThread.wait_for_state_change;
var
  value: string;
  waitRes: boolean;
begin
  // maybe after ORB shutdown
  if ORB_Instance <> nil then
    value := ORB_Instance.get_properties().get_Property('mtdorb.orb.thread_pool_timeout');
  if (FState = wtsIdle) and (value <> '') then
    waitRes := FStateSem.wait(StrToInt(value) * 1000)
  else
    waitRes := FStateSem.wait();
  if not waitRes and (value <> '') and FTP.remove_thread(Self) then begin
    FState := wtsStop;
    register_operation(nil);
  end;
end;

{ TThreadPool }

constructor TThreadPool.Create(AMax, AMaxIdle, AMinIdle: cardinal);
begin
  //FTPLock := TMutex.Create;
  FBusyThreads := TInterfaceList.Create;//TList.Create;
  FIdleThreads := TInterfaceList.Create;//TList.Create;
  FMonitor := TThreadMonitor.Create;
  FMax := AMax;
  FMaxIdle := AMaxIdle;
  FMinIdle := AMinIdle;
end;

destructor TThreadPool.Destroy;
var
  i: integer;
begin
  //terminate all threads
  for i := 0 to FIdleThreads.Count - 1 do
    IWorkerThread(FIdleThreads.Items[i]).Terminate(nil);
  for i := 0 to FBusyThreads.Count - 1 do
    IWorkerThread(FBusyThreads.Items[i]).Terminate(nil);

  //FIdleThreads.Free;
  //FBusyThreads.Free;
  //FTPLock.Free;
  FMonitor.Free;
  inherited;
end;

// returns - a idle thread or creates a new thread if needed
// - or NIL if limits are exhausted
// removes threads from the idle list
function TThreadPool.get_idle_thread: IWorkerThread;
var
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FMonitor);
  {FTPLock.lock();
  try}
    // no idle threads
    if FIdleThreads.Count = 0 then begin
      // limit exhausted
      if FCountAll >= FMax then Exit;
      result := TWorkerThread.Create(Self);
      if FOp <> nil then result.register_operation(FOp.copy());
      FCountAll := Succ(FCountAll);
      result.Start();
    end
    else begin
      result := IWorkerThread(FIdleThreads.Items[0]);
      FIdleThreads.Delete(0);
    end;
    result.set_state(wtsTransition);
  {finally
    FTPLock.unlock();
  end; { try/finally }
end;

function TThreadPool.getInputMC: IMsgChannel;//TMsgChannel;
begin
  result := FInputMC;
end;

function TThreadPool.get_operation: IOperation;
begin
  result := FOp;
end;

function TThreadPool.get_thread_pool_manager: IThreadPoolManager;
begin
  result := FTpm;
end;

function TThreadPool.new_idle_thread: IWorkerThread;
var
  sync: ISynchronized;
begin
  result := TWorkerThread.Create(Self);
  sync := TSynchronized.Create(FMonitor);
  {FTPLock.lock();
  try}
    //result._AddRef;
    result.SetNo(FIdleThreads.Add(result));
    result.set_state(wtsIdle);
    FCountAll := Succ(FCountAll);
  {finally
    FTPLock.unlock();
  end; { try/finally }
  result.start();
end;

procedure TThreadPool.put_msg(nextOpId: TOperationKind; msgType: IMsgType);
begin
  //FOp.put_msg(nextOpId, msgType);
  if FInputMC <> nil then
    FInputMC.put_msg(nextOpId, msgType);
end;

procedure TThreadPool.registerInputMC(mc: TMsgChannel);
begin
  FInputMC := mc;
  FInputMC.set_thread_pool(Self);
end;

function TThreadPool.remove_thread(kt: IWorkerThread): boolean;
var
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FMonitor);
  {FTPLock.lock();
  try}
    result := kt.get_state() <> wtsTransition;
    if result then begin
      FIdleThreads.Remove(kt);//Remove(Pointer(kt));
      FBusyThreads.Remove(kt);//Remove(Pointer(kt));
    {case kt.State of
      wtsIdle: FIdleThreads.Remove(Pointer(kt));//Delete(kt.no);
      wtsBusy: FBusyThreads.Remove(Pointer(kt));//Delete(kt.no)
    end;}
    if FCountAll > 0 then
      FCountAll := Pred(FCountAll);
    end;
  {finally
    FTPLock.unlock();
  end; { try/finally }
end;

procedure TThreadPool.set_operation(op: IOperation);
begin
  FOp := op;
end;

procedure TThreadPool.set_thread_pool_manager(tpm: IThreadPoolManager);
begin
  FTpm := tpm;
end;

// fire up all threads
procedure TThreadPool.start_threads(arg: pointer);
var
  kt: IWorkerThread;
  i: integer;
begin
  for i := 1 to FMinIdle do begin
    kt := new_idle_thread();
    if get_operation() <> nil then kt.register_operation(get_operation().copy());
  end;
end;

procedure TThreadPool.mark_busy(wt: IWorkerThread);
var
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FMonitor);
  {FTPLock.lock();
  try}
    if wt.get_state() = wtsIdle then begin
      FIdleThreads.Remove(wt);//(Pointer(wt));// Delete(wt.no);
      //wt._Release;
    end;
    {else} //wt._AddRef;
    wt.SetNo(FBusyThreads.Add(wt));
    wt.set_state(wtsBusy);
  {finally
    FTPLock.unlock();
  end; { try/finally }
end;

procedure TThreadPool.mark_idle(wt: IWorkerThread);
{var
  wt_dummy: IWorkerThread;}
var
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FMonitor);
  //if (FInputMC <> nil) and FInputMC.check_msg(wt) then Exit;
  {FTPLock.lock();
  try}
    if wt.get_state() = wtsBusy then begin
      FBusyThreads.Remove(wt);//(Pointer(wt));// Delete(wt.no);
      //wt._Release;
    end;
    {else} //wt._AddRef;
    wt.SetNo(FIdleThreads.Add(wt));//(Pointer(wt));
    wt.set_state(wtsIdle);
    if FInputMC <> nil then
      FInputMC.check_msg({wt_dummy});
    if FBusyThreads.Count = 0 then
      FMonitor.notifyAll();
  {finally
    FTPLock.unlock();
  end; { try/finally }
end;

function TThreadPool.optype: integer;
begin
  result := FOpType;
end;

procedure TThreadPool.clear;
var
  wt: IWorkerThread;
  //cnt: integer;
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FMonitor);
  FOp := nil;
  while FBusyThreads.Count > 0 do
    FMonitor.wait();
  while FIdleThreads.Count > 0 do begin
    wt := IWorkerThread(FIdleThreads.Last);
    wt.register_operation(nil);
    wt.set_state(wtsStop);
    wt.post_state_change();
    wt.wait();
    remove_thread(wt);
    wt := nil;
  end;
  FInputMC := nil;
end;

{ TThreadPoolManager }

constructor TThreadPoolManager.Create();
begin
  FUserCnt := 64;
  FTp := TStringList.Create;
end;

destructor TThreadPoolManager.Destroy;
var
  i: integer;
  tp: IThreadPool;
begin
  for i := 0 to FTp.Count - 1 do begin
    tp := IThreadPool(Pointer(FTp.Objects[i]));
    tp._Release;
    //tp.clear(); // do not wait 
    tp := nil;
  end;
  FTp.Free;
  inherited;
end;

function TThreadPoolManager.get_idle_thread(
  const opId: TOperationKind): IWorkerThread;
begin
  result := get_thread_pool(opId).get_idle_thread();
end;

function TThreadPoolManager.get_operation(opId: TOperationKind): IOperation;
begin
  result := get_thread_pool(opId).get_operation();
end;

function TThreadPoolManager.get_thread_pool(
  const opId: TOperationKind): IThreadPool;
begin
  result := IThreadPool(Pointer(FTp.Objects[FTp.IndexOf(IntToStr(Ord(opId)))]));
end;

procedure TThreadPoolManager.put_msg(nextOpId: TOperationKind;
  msgType: IMsgType);
begin
  get_thread_pool(nextOpId).put_msg(nextOpId, msgType);
end;

function TThreadPoolManager.register_new_tp(tp: IThreadPool): _ulong;
begin
  tp._AddRef;
  FTp.AddObject(IntToStr(FUserCnt), Pointer(tp));
  tp.set_thread_pool_manager(Self as IThreadPoolManager);
  FUserCnt := Succ(FUserCnt);
  result := FUserCnt;
end;

procedure TThreadPoolManager.register_tp(const opId: TOperationKind;
  tp: IThreadPool);
var
  idx: integer;
begin
  idx := FTp.IndexOf(IntToStr(Ord(opId)));
  tp._AddRef;
  if idx <> -1 then begin
    IUnknown(Pointer(FTp.Objects[idx]))._Release;
    FTp.Objects[idx] := Pointer(tp);
  end
  else FTp.AddObject(IntToStr(Ord(opId)), Pointer(tp));
end;

procedure TThreadPoolManager.shutdown;
var
  i: integer;
  tp: IThreadPool;
begin
  for i := 0 to FTp.Count - 1 do begin
    tp := IThreadPool(Pointer(FTp.Objects[i]));
    //tp._AddRef();
    tp.clear();
    tp := nil;
  end;
end;

{ TMTDispatcher }

function TMTDispatcher.copy: IOperation;
begin
  result := TMTDispatcher.Create(Self);
end;

constructor TMTDispatcher.Create;
begin
  inherited;
  FInfo.set_OP_id(odORB);
end;

constructor TMTDispatcher.Create(op: IOperation);
begin
  inherited;
  FInfo.set_OP_id(odORB);
end;

procedure TMTDispatcher.process(msg: IMsgType);
var
  msgRec: IOrbMsg;
begin
  msgRec := IOrbMsg(msg.Data());
  case msgRec.orbEvent() of
    omeKillConn:
      begin
        Assert(msgRec.connection().state = esTerminated);
      end;
    omeCloseConn: begin
      msgRec.callback().giop_callback(msgRec.connection(), Closed);
    end;
  end;
  msgRec._Release();
end;

initialization
finalization
  tpmVar := nil;
{$ELSE}
implementation
{$ENDIF}
end.
