// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit osnet;

interface

uses SysUtils, orbtypes, {$IFDEF WIN32}WinSock, Windows{$ENDIF}{$IFDEF LINUX}Libc{$ENDIF};

{$DEFINE TCP_NODELAY}

type
  TOSNet = class(TObject)
  public
    class procedure sockinit();
    class procedure sockfinal();
    class procedure seterrno();
    class procedure sockblock(s: TSocket; val: Boolean);
    class function sockndelay(s: TSocket; val: Boolean): integer;
    class function sockclose(s: TSocket): long;
    class function sockwrite(s: TSocket; var Buffer; len: _ulong): integer;
    class function sockread(s: TSocket; var Buffer; len: _ulong): integer;
    class procedure sockreuse(s: TSocket; val: Boolean);
  end;

{$IFDEF WIN32}
function errno(): long;
{$ENDIF}

implementation

uses throw, except_int;

threadvar
  errno_var: long;

{$IFDEF WIN32}
function errno(): long;
begin
  result := errno_var;
end;
{$ENDIF}

{ TOSNet }

class procedure TOSNet.seterrno;
{$IFDEF WIN32}
var
  e: longint;
{$ENDIF}
begin
{$IFDEF WIN32}
  e := WSAGetLastError();
  case e of
    WSAEINTR: errno_var := EINTR;
    WSAEWOULDBLOCK: errno_var := EWOULDBLOCK;
    WSAECONNRESET: errno_var := ECONNRESET;
    WSAESHUTDOWN: errno_var := ESHUTDOWN;
    WSAECONNABORTED: errno_var := ECONNABORTED;
    else
      errno_var := -1;
  end;
{$ENDIF}
end;

class procedure TOSNet.sockblock(s: TSocket; val: Boolean);
var
{$IFDEF WIN32}
  tmp: integer;
{$ENDIF}
{$IFDEF LINUX}
  flags: long;
{$ENDIF}
begin
{$IFDEF WIN32}
  if val then
    tmp := 0
  else
    tmp := 1;
  Assert(winsock.ioctlsocket(s, FIONBIO, tmp) = 0);
{$ENDIF}
{$IFDEF LINUX}
  flags := fcntl(s, F_GETFL, 0);
  Assert(flags <> -1);
  if val then
    flags := flags and not O_NONBLOCK
  else
    flags := flags or O_NONBLOCK;
  fcntl(s, F_SETFL, flags);
{$ENDIF}
end;

class function TOSNet.sockclose(s: TSocket): long;
begin
{$IFDEF WIN32}
  WSASetLastError(0);
  result := winsock.closesocket(s);
  if result = 0 then exit;
  seterrno();
  result := -1;
{$ENDIF}
{$IFDEF LINUX}
  result := __close(s);
{$ENDIF}
end;

class procedure TOSNet.sockfinal;
{$IFDEF LINUX}
var
  actPIPE: TSigAction;
{$ENDIF}
begin
{$IFDEF MSWINDOWS}
  WSACleanup();
{$ENDIF}
{$IFDEF LINUX}
  actPIPE.__sigaction_handler := TSignalHandler(SIG_DFL);
  sigemptyset(actPIPE.sa_mask);
  actPIPE.sa_flags := 0;
  sigaction(SIGPIPE, @actPIPE, nil);
{$ENDIF}
end;

class procedure TOSNet.sockinit;
var
{$IFDEF WIN32}
  ver: WORD;
  err: integer;
  dat: TWSADATA;
{$ENDIF}
{$IFDEF LINUX}
  actPIPE: TSigAction;
{$ENDIF}
begin
{$IFDEF WIN32}
  ver := MakeWord(2,0);
  err := WSAStartup(ver, dat);
  if err <> 0 then
    dorb_throw(st_COMM_FAILURE, MinorWSAStartup, COMPLETED_NO);
{$ENDIF}
{$IFDEF LINUX}
  //when peer closes a unix socket we receive SIGPIPE
  actPIPE.__sigaction_handler := TSignalHandler(SIG_IGN);
  sigemptyset(actPIPE.sa_mask);
  actPIPE.sa_flags := 0;
  sigaction(SIGPIPE, @actPIPE, nil);
  //signal(SIGPIPE, TSignalHandler(SIG_IGN));
{$ENDIF}
end;

class function TOSNet.sockndelay(s: TSocket; val: Boolean): integer;
{$IFDEF TCP_NODELAY}
var
  tmp: integer;
{$ENDIF}
begin
{$IFDEF TCP_NODELAY}
  tmp := Integer(val);
  result := setsockopt(s, IPPROTO_TCP, TCP_NODELAY, PChar(@tmp), sizeof(tmp));
  if result = 0 then exit;
  seterrno();
{$ELSE}
  result := 0;
{$ENDIF}
end;

class function TOSNet.sockread(s: TSocket; var Buffer; len: _ulong): integer;
begin
{$IFDEF WIN32}
  WSASetLastError(0);
  result := winsock.recv(s,buffer,len,0);
  if result <> SOCKET_ERROR then
    exit;
  seterrno();
{$ENDIF}
{$IFDEF LINUX}
  result := recv(s,buffer,len,0);
{$ENDIF}
end;

class procedure TOSNet.sockreuse(s: TSocket; val: Boolean);
var
  _val : longint;
begin
  if val then
    _val := 1
  else
    _val := 0;
  assert(setsockopt (s, SOL_SOCKET, SO_REUSEADDR, PChar(@_val), sizeof(_val))=0);
end;

class function TOSNet.sockwrite(s: TSocket; var Buffer; len: _ulong): integer;
begin
{$IFDEF WIN32}
  WSASetLastError(0);
  // Limit the maximum size of a message, in order to avoid
  // a performance problem on Win32.
  //
  // Note: The maximum size must be less than the socket's
  // send buffer size set by the constructor.
  //
  // TODO: This is a total hack!
  // Thanks guys from OOC
  if (len > 65535) then
    len := 65535;
  result := winsock.send(s, buffer, len, 0);
  if result <> SOCKET_ERROR then
    exit;
  seterrno();
{$ENDIF}
{$IFDEF LINUX}
  result := send(s, buffer, len, 0);
{$ENDIF}
end;

end.
